package field

import (
	"context"

	"git.perx.ru/perxis/perxis-go/pkg/errors"
)

type Decoder interface {
	Decode(ctx context.Context, field *Field, v interface{}) (interface{}, error)
}

type Encoder interface {
	Encode(ctx context.Context, field *Field, v interface{}) (interface{}, error)
}

// NonStrictConverter определяет метод для преобразования данных в нестрогом режиме.
type NonStrictConverter interface {
	NonStrictConvert(ctx context.Context, field *Field, v interface{}) interface{}
}

type EncodeOptions struct {
	NonStrictMode bool
}

type EncodeOption func(opts *EncodeOptions)

// NonStrictMode указывает, что декодирование необходимо провести в нестрогом режиме.
func NonStrictMode() EncodeOption {
	return func(opts *EncodeOptions) {
		opts.NonStrictMode = true
	}
}

func NewEncodeOptions(opt ...EncodeOption) *EncodeOptions {
	opts := &EncodeOptions{}
	for _, o := range opt {
		o(opts)
	}
	return opts
}

func Decode(ctx context.Context, w Walker, v interface{}, opts ...EncodeOption) (interface{}, error) {
	opt := NewEncodeOptions(opts...)
	var walkOpts []WalkOption
	if opt.NonStrictMode {
		walkOpts = append(walkOpts, WalkNonStrict())
	}

	val, _, err := w.Walk(ctx, v, func(ctx context.Context, f *Field, v interface{}) (res WalkFuncResult, err error) {
		decoder, ok := f.GetType().(Decoder)
		if !ok {
			res.Value = v
			return res, nil
		}

		if opt.NonStrictMode {
			if converter, hasConv := f.GetType().(NonStrictConverter); hasConv {
				v = converter.NonStrictConvert(ctx, f, v)
			}
		}

		if v, err = decoder.Decode(ctx, f, v); err != nil {
			return
		}
		res.Value = v
		res.Changed = true
		return
	}, walkOpts...)

	if err != nil {
		return nil, errors.Wrap(err, "decode error")
	}

	return val, nil
}

func Encode(ctx context.Context, w Walker, v interface{}) (interface{}, error) {
	val, _, err := w.Walk(ctx, v, func(ctx context.Context, f *Field, v interface{}) (res WalkFuncResult, err error) {
		if encode, ok := f.GetType().(Encoder); ok {
			if v, err = encode.Encode(ctx, f, v); err != nil {
				return
			}
			res.Value = v
			res.Changed = true
			return
		}
		res.Value = v
		return
	})
	if err != nil {
		return nil, errors.Wrap(err, "encode error")
	}
	return val, nil

}
