package template

import (
	"context"
	"errors"
	"testing"

	"git.perx.ru/perxis/perxis-go/pkg/collections"
	colsmocks "git.perx.ru/perxis/perxis-go/pkg/collections/mocks"
	"git.perx.ru/perxis/perxis-go/pkg/content"
	"git.perx.ru/perxis/perxis-go/pkg/environments"
	envsmocks "git.perx.ru/perxis/perxis-go/pkg/environments/mocks"
	"git.perx.ru/perxis/perxis-go/pkg/items"
	mocksitems "git.perx.ru/perxis/perxis-go/pkg/items/mocks"
	"git.perx.ru/perxis/perxis-go/pkg/spaces"
	spsmocks "git.perx.ru/perxis/perxis-go/pkg/spaces/mocks"
	"github.com/stretchr/testify/assert"
)

func TestBuilder_Execute(t *testing.T) {
	tests := []struct {
		name    string
		SpaceID string
		EnvID   string
		CollID  string
		str     string
		data    any
		want    any
		wantErr bool

		getCnt func(t *testing.T) *content.Content
	}{
		{name: "error", str: "hello {{ .a }}", data: "world", want: "", wantErr: true},
		{name: "empty", str: "", data: "", want: "", wantErr: false},
		{name: "#1", str: "hello {{ . }}", data: "world", want: "hello world", wantErr: false},
		{name: "#2", str: "{{ . }}", data: "world", want: "world", wantErr: false},
		{name: "#3 ", str: "", data: "world", want: "", wantErr: false},
		{name: "#4 ", str: "hello", data: "world", want: "hello", wantErr: false},
		{
			name:    "lookup",
			SpaceID: "space",
			EnvID:   "env",
			str:     "hello, {{ lookup \"secrets.dev.key\" }}",
			data:    "",
			want:    "hello, Luk",
			wantErr: false,
			getCnt: func(t *testing.T) *content.Content {
				itemsSvc := mocksitems.NewItems(t)
				itemsSvc.On("Get", context.Background(), "space", "env", "secrets", "dev").
					Return(&items.Item{
						ID:           "dev",
						SpaceID:      "space",
						EnvID:        "env",
						CollectionID: "secrets",
						Data: map[string]interface{}{
							"id":  "dev",
							"key": "Luk",
						},
					}, nil).Once()
				return &content.Content{Items: itemsSvc}
			},
		},
		{
			name:    "lookup with slice",
			SpaceID: "space",
			EnvID:   "env",
			str:     "numbers {{ lookup \"secrets.dev.slice\" }}",
			data:    "",
			want:    "numbers [1 2 3]",
			wantErr: false,
			getCnt: func(t *testing.T) *content.Content {
				itemsSvc := mocksitems.NewItems(t)
				itemsSvc.On("Get", context.Background(), "space", "env", "secrets", "dev").
					Return(&items.Item{
						ID:           "dev",
						SpaceID:      "space",
						EnvID:        "env",
						CollectionID: "secrets",
						Data: map[string]interface{}{
							"id":    "dev",
							"slice": []int{1, 2, 3},
						},
					}, nil).Once()
				return &content.Content{Items: itemsSvc}
			},
		},
		{
			name:    "lookup with empty Data",
			SpaceID: "space",
			EnvID:   "env",
			str:     "numbers {{ lookup \"secrets.dev.slice\" }}",
			data:    "",
			want:    "numbers <no value>",
			wantErr: false,
			getCnt: func(t *testing.T) *content.Content {
				itemsSvc := mocksitems.NewItems(t)
				itemsSvc.On("Get", context.Background(), "space", "env", "secrets", "dev").
					Return(&items.Item{
						ID:           "dev",
						SpaceID:      "space",
						EnvID:        "env",
						CollectionID: "secrets",
						Data:         map[string]interface{}{},
					}, nil).Once()
				return &content.Content{Items: itemsSvc}
			},
		},
		{
			name:    "lookup with incorrect field",
			SpaceID: "space",
			EnvID:   "env",
			str:     "hello {{ lookup \"secrets.dev.incorrect\" }}",
			data:    "",
			want:    "hello <no value>",
			wantErr: false,
			getCnt: func(t *testing.T) *content.Content {
				itemsSvc := mocksitems.NewItems(t)
				itemsSvc.On("Get", context.Background(), "space", "env", "secrets", "dev").
					Return(&items.Item{
						ID:           "dev",
						SpaceID:      "space",
						EnvID:        "env",
						CollectionID: "secrets",
						Data: map[string]interface{}{
							"id":  "dev",
							"key": "1234",
						},
					}, nil).Once()
				return &content.Content{Items: itemsSvc}
			},
		},
		{
			name:    "lookup not found",
			SpaceID: "space",
			EnvID:   "env",
			str:     "hello {{ lookup \"secrets.prod.pass\" }}",
			data:    "",
			want:    "",
			wantErr: true,
			getCnt: func(t *testing.T) *content.Content {
				itemsSvc := mocksitems.NewItems(t)
				itemsSvc.On("Get", context.Background(), "space", "env", "secrets", "prod").
					Return(nil, errors.New("not found")).Once()
				return &content.Content{Items: itemsSvc}
			},
		},
		{name: "lookup without itemID", SpaceID: "space", EnvID: "env", str: "hello {{ lookup \"secrets.pass\" }}", data: "", want: "", wantErr: true},
		{name: "system#1", SpaceID: "space", str: "hello {{ system.SpaceID }}", data: "", want: "hello space", wantErr: false},
		{name: "system space", SpaceID: "space", getCnt: func(t *testing.T) *content.Content {
			spsSvc := spsmocks.NewSpaces(t)
			spsSvc.On("Get", context.Background(), "space").Return(&spaces.Space{Description: "description"}, nil).Once()
			return &content.Content{Spaces: spsSvc}
		}, str: "{{ system.Space.Description }}", want: "description", wantErr: false},
		{name: "system environment", SpaceID: "space", EnvID: "env", getCnt: func(t *testing.T) *content.Content {
			envsSvc := envsmocks.NewEnvironments(t)
			envsSvc.On("Get", context.Background(), "space", "env").Return(&environments.Environment{Aliases: []string{"master"}}, nil).Once()
			return &content.Content{Environments: envsSvc}
		}, str: "{{ system.Environment.Aliases }}", want: "[master]", wantErr: false},
		{
			name:    "system collection",
			SpaceID: "space",
			EnvID:   "env",
			CollID:  "col",
			getCnt: func(t *testing.T) *content.Content {
				collsSvc := colsmocks.NewCollections(t)
				collsSvc.On("Get", context.Background(), "space", "env", "col").
					Return(&collections.Collection{Name: "cars"}, nil).Once()
				return &content.Content{Collections: collsSvc}
			},
			str:     "{{ system.Collection.Name }}",
			want:    "cars",
			wantErr: false,
		},
		{name: "system without account", SpaceID: "space", str: "hello {{ system.Organization.Name }}", want: "", wantErr: true},
	}
	for _, tt := range tests {
		t.Run(tt.name, func(t *testing.T) {
			var cnt *content.Content
			if tt.getCnt != nil {
				cnt = tt.getCnt(t)
			}

			b := &Builder{
				ctx:     context.Background(),
				cnt:     cnt,
				SpaceID: tt.SpaceID,
				EnvID:   tt.EnvID,
				CollID:  tt.CollID,
			}

			got, err := b.Execute(tt.str, tt.data)
			if tt.wantErr == true {
				assert.Error(t, err)
			} else {
				assert.NoError(t, err)
			}
			assert.Equal(t, tt.want, got)
		})
	}
}

func TestBuilder_ExecuteList(t *testing.T) {
	tests := []struct {
		name    string
		SpaceID string
		EnvID   string
		str     []string
		data    any
		want    []string
		wantErr bool

		itemsCall func(itemsSvc *mocksitems.Items)
	}{
		{name: "error", SpaceID: "space", EnvID: "env", str: []string{"hello { . }}", "go {{ . }"}, data: "world", want: []string{}, wantErr: true},
		{name: "empty", SpaceID: "space", EnvID: "env", str: []string{""}, data: "world", want: []string{""}, wantErr: false},
		{name: "#1", SpaceID: "space", EnvID: "env", str: []string{"hello {{ . }}", "go {{ . }}"}, data: "world", want: []string{"hello world", "go world"}, wantErr: false},
		{name: "#2", SpaceID: "space", EnvID: "env", str: []string{"{{ . }}"}, data: "world", want: []string{"world"}, wantErr: false},
		{name: "#3 ", SpaceID: "space", EnvID: "env", str: []string{""}, data: "world", want: []string{""}, wantErr: false},
		{name: "#4 ", SpaceID: "space", EnvID: "env", str: []string{"hello"}, data: "world", want: []string{"hello"}, wantErr: false},
		{name: "lookup", SpaceID: "space", EnvID: "env", str: []string{"hello {{ lookup \"secrets.dev.key\" }}"}, data: "", want: []string{"hello 1234"}, wantErr: false, itemsCall: func(itemsSvc *mocksitems.Items) {
			itemsSvc.On("Get", context.Background(), "space", "env", "secrets", "dev").Return(&items.Item{
				ID:           "dev",
				SpaceID:      "space",
				EnvID:        "env",
				CollectionID: "secrets",
				Data: map[string]interface{}{
					"id":  "dev",
					"key": "1234",
				},
			}, nil).Once()
		}},
		{name: "lookup with incorrect field", SpaceID: "space", EnvID: "env", str: []string{"hello {{ lookup \"secrets.dev.incorrect\" }}"}, data: "", want: []string{"hello <no value>"}, wantErr: false, itemsCall: func(itemsSvc *mocksitems.Items) {
			itemsSvc.On("Get", context.Background(), "space", "env", "secrets", "dev").Return(&items.Item{
				ID:           "dev",
				SpaceID:      "space",
				EnvID:        "env",
				CollectionID: "secrets",
				Data: map[string]interface{}{
					"id":  "dev",
					"key": "1234",
				},
			}, nil).Once()
		}},
		{name: "system ", SpaceID: "space", EnvID: "env", str: []string{"hello {{ system.SpaceID }}"}, data: "", want: []string{"hello space"}, wantErr: false},
	}
	for _, tt := range tests {
		t.Run(tt.name, func(t *testing.T) {
			itemsSvc := mocksitems.NewItems(t)
			if tt.itemsCall != nil {
				tt.itemsCall(itemsSvc)
			}
			b := &Builder{
				ctx:     context.Background(),
				cnt:     &content.Content{Items: itemsSvc},
				SpaceID: tt.SpaceID,
				EnvID:   tt.EnvID,
			}

			got, err := b.ExecuteList(tt.str, tt.data)
			if tt.wantErr == true {
				assert.Error(t, err)
			} else {
				assert.NoError(t, err)
			}
			assert.Equal(t, tt.want, got)
		})
	}
}

func TestBuilder_ExecuteMap(t *testing.T) {
	tests := []struct {
		name    string
		SpaceID string
		EnvID   string
		str     map[string]interface{}
		data    any
		want    map[string]interface{}
		wantErr bool

		itemsCall func(itemsSvc *mocksitems.Items)
	}{
		{name: "error", SpaceID: "space", EnvID: "env", str: map[string]interface{}{"hello": "{{ . }"}, data: "world", want: nil, wantErr: true},
		{name: "empty", SpaceID: "space", EnvID: "env", str: map[string]interface{}{}, data: "", want: map[string]interface{}{}, wantErr: false},
		{name: "#1", SpaceID: "space", EnvID: "env", str: map[string]interface{}{"hello": "{{ . }}"}, data: "world", want: map[string]interface{}{"hello": "world"}, wantErr: false},
		{name: "#2", SpaceID: "space", EnvID: "env", str: map[string]interface{}{"hello": "{{ . }}", "go": "{{ . }}"}, data: "world", want: map[string]interface{}{"hello": "world", "go": "world"}, wantErr: false},
		{name: "#3 ", SpaceID: "space", EnvID: "env", str: map[string]interface{}{}, data: "world", want: map[string]interface{}{}, wantErr: false},
		{name: "#4 ", SpaceID: "space", EnvID: "env", str: map[string]interface{}{"a": "b"}, data: "world", want: map[string]interface{}{"a": "b"}, wantErr: false},
		{name: "lookup ", SpaceID: "space", EnvID: "env", str: map[string]interface{}{"hello": "{{ lookup \"secrets.dev.key\" }}"}, data: "", want: map[string]interface{}{"hello": "1234"}, wantErr: false, itemsCall: func(itemsSvc *mocksitems.Items) {
			itemsSvc.On("Get", context.Background(), "space", "env", "secrets", "dev").Return(&items.Item{
				ID:           "dev",
				SpaceID:      "space",
				EnvID:        "env",
				CollectionID: "secrets",
				Data: map[string]interface{}{
					"id":  "dev",
					"key": "1234",
				},
			}, nil).Once()
		}},
		{name: "lookup with incorrect field", SpaceID: "space", EnvID: "env", str: map[string]interface{}{"hello": "{{ lookup \"secrets.dev.incorrect\" }}"}, data: "", want: map[string]interface{}{"hello": "<no value>"}, wantErr: false, itemsCall: func(itemsSvc *mocksitems.Items) {
			itemsSvc.On("Get", context.Background(), "space", "env", "secrets", "dev").Return(&items.Item{
				ID:           "dev",
				SpaceID:      "space",
				EnvID:        "env",
				CollectionID: "secrets",
				Data: map[string]interface{}{
					"id":  "dev",
					"key": "1234",
				},
			}, nil).Once()
		}},
		{name: "system ", SpaceID: "space", EnvID: "env", str: map[string]interface{}{"hello": "{{ system.SpaceID }}"}, data: "", want: map[string]interface{}{"hello": "space"}, wantErr: false},
	}
	for _, tt := range tests {
		t.Run(tt.name, func(t *testing.T) {
			itemsSvc := mocksitems.NewItems(t)
			if tt.itemsCall != nil {
				tt.itemsCall(itemsSvc)
			}
			b := &Builder{
				ctx:     context.Background(),
				cnt:     &content.Content{Items: itemsSvc},
				SpaceID: tt.SpaceID,
				EnvID:   tt.EnvID,
			}

			got, err := b.ExecuteMap(tt.str, tt.data)
			if tt.wantErr == true {
				assert.Error(t, err)
			} else {
				assert.NoError(t, err)
			}
			assert.Equal(t, tt.want, got)
		})
	}
}
